@@ -1,11 +1,10 @@
-from .backend import np
+import copy
 import inspect
+import weakref
+from enum import Enum
+import sys
 
-# TOOO:
-# add switch for cpu vs cupy
-# add channel family support
-# add system commands
-# add debug support
+from .backend import np
 
 
 class KGSym(str):
@@ -16,7 +15,8 @@ class KGSym(str):
     def __hash__(self):
         return super().__hash__()
 
-class KGFn():
+
+class KGFn:
     def __init__(self, a, args, arity):
         self.a = a
         self.args = args
@@ -27,20 +27,28 @@ class KGFn():
             return ":nilad"
         elif self.arity == 1:
             return ":monad"
-        return ":dyad"
+        elif self.arity == 2:
+            return ":dyad"
+        return ":triad"
+
+    def is_op(self):
+        return isinstance(self.a,KGOp)
+
+    def is_adverb_chain(self):
+        return (isinstance(self.a,list) and isinstance(self.a[0], KGAdverb))
 
 
 class KGCall(KGFn):
     pass
 
 
-class KGOp():
+class KGOp:
     def __init__(self, a, arity):
         self.a = a
         self.arity = arity
 
 
-class KGAdverb():
+class KGAdverb:
     def __init__(self, a, arity):
         self.a = a
         self.arity = arity
@@ -54,7 +62,7 @@ class KGCond(list):
     pass
 
 
-class KGLambda():
+class KGLambda:
     """
     KGLambda wraps a lambda and make it available to Klong, allowing for direct
     integration of python functions in Klong.
@@ -87,13 +95,38 @@ class KGLambda():
         return len(self.args) - 1 if self.provide_klong else len(self.args)
 
 
-class UnresolvedArgument(Exception):
-    pass
+class KGChannelDir(Enum):
+    INPUT=1
+    OUTPUT=2
+
+
+class KGChannel:
+    class FileHandler:
+        def __init__(self, raw, parent):
+            self._raw = raw
+            self._ref = weakref.ref(parent, self.close)
+
+        def close(self, *args):
+            self._raw.close()
+
+    def __init__(self, raw, channel_dir):
+        self.channel_dir = channel_dir
+        self.raw = raw
+        self._fh = KGChannel.FileHandler(raw, self)
+        self.at_eof = False
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, ext_type, exc_value, traceback):
+        self._fh.close()
+
 
 
 class RangeError(Exception):
     def __init__(self, i):
         self.i = i
+        super().__init__()
 
 
 reserved_fn_args = ['x','y','z']
@@ -101,12 +134,8 @@ reserved_fn_symbols = {KGSym(n) for n in reserved_fn_args}
 reserved_fn_symbol_map = {n:KGSym(n) for n in reserved_fn_args}
 
 
-def die(m=None):
-    raise RuntimeError(m)
-
-
 def is_list(x):
-    return isinstance(x,(list, np.ndarray))
+    return isinstance(x,list) or (isinstance(x, np.ndarray) and len(x.shape) > 0)
 
 
 def is_iterable(x):
@@ -175,14 +204,14 @@ def cpeek2(t,i):
 
 class UnexpectedChar(Exception):
     def __init__(self, t, i, c):
-        self.t = t
-        self.i = i
-        self.c = c
+        super().__init__(f"t: {t[i-10:i+10]} pos: {i} char: {c}")
+
 
 class UnexpectedEOF(Exception):
     def __init__(self, t, i):
         self.t = t
         self.i = i
+        super().__init__(f"t: {t[i-10:]} pos: {i}")
 
 
 def cexpect(t, i, c):
@@ -285,6 +314,8 @@ def kg_truth(x):
     return x*1
 
 
+# TODO: can we just transform chars to ints so that CuPy works?
+#       we'll need to reassemble the strinsg, so pros/cons.
 def str_to_chr_arr(s):
     return np.asarray([KGChar(x) for x in s],dtype=object)
 
@@ -295,8 +326,12 @@ def read_num(t, i=0):
     if t[i] == '-':
         i += 1
     while i < len(t):
-        if t[i] == '.' or t[i] == 'e':
+        if t[i] == '.':
             use_float = True
+        elif t[i] == 'e':
+            use_float = True
+            if cmatch(t,i+1,'-'):
+                i += 2
         elif not t[i].isnumeric():
             break
         i += 1
@@ -310,12 +345,18 @@ def read_char(t, i):
     return i+1, KGChar(t[i])
 
 
-def read_sym(t, i=0):
+def read_sym(t, i=0, module=None):
     p = i
     while i < len(t) and is_symbolic(t[i]):
         i += 1
     x = t[p:i]
-    return i, reserved_fn_symbol_map.get(x) or KGSym(x)
+    return i, reserved_fn_symbol_map.get(x) or KGSym(x if x.startswith('.') or module is None else f"{x}`{module}")
+
+
+def read_op(t, i=0):
+    if cmatch2(t,i,'\\','~') or cmatch2(t,i,'\\','*'):
+        return i+2,KGOp(t[i:i+2],arity=0)
+    return i+1,KGOp(t[i:i+1],arity=0)
 
 
 def read_shifted_comment(t, i=0):
@@ -347,19 +388,26 @@ def read_sys_comment(t,i,a):
 
     """
     try:
-        return i + t[i:].index(a) + len(a)
+        j = t[i:].index(a)
+        while t[i+j+1:].startswith(a):
+            j += 1
+        return i + j + len(a)
     except ValueError:
         return RuntimeError("end of comment not found")
 
-
-def skip_space(t, i=0):
-    while i < len(t) and t[i].isspace():
+def skip_space(t, i=0, ignore_newline=False):
+    """
+        NOTE: a newline character translates to a semicolon in Klong,
+        except in functions, dictionaries, conditional expressions,
+        and lists. So
+    """
+    while i < len(t) and (t[i].isspace() and (ignore_newline or t[i] != '\n')):
         i += 1
     return i
 
 
-def skip(t, i=0):
-    i = skip_space(t,i)
+def skip(t, i=0, ignore_newline=False):
+    i = skip_space(t,i,ignore_newline=ignore_newline)
     if cmatch2(t, i, ':', '"'):
         i = read_shifted_comment(t, i+2)
         i = skip(t, i)
@@ -379,7 +427,7 @@ def cast_malformed_array(arr):
     return np.asarray(r,dtype=object)
 
 
-def read_list(t, delim, i=0):
+def read_list(t, delim, i=0, module=None):
     """
 
         # A list is any number of class lexemes (or lists) delimited by
@@ -388,17 +436,20 @@ def read_list(t, delim, i=0):
         L := '[' (C|L)* ']'
 
     """
-
     arr = []
     obj_arr = False
+    i = skip(t,i,ignore_newline=True)
     while not cmatch(t,i,delim) and i < len(t):
         # TODO: make cleaner: kind of a hack to pass in read_neg but
         #       we can knowingly read neg numbers in list context
-        i, q = kg_token(t, i, read_neg=True)
+        i, q = kg_read(t, i, read_neg=True, ignore_newline=True, module=module)
+        if q is None:
+            break
         if safe_eq(q, '['):
-            i,q = read_list(t, ']', i)
+            i,q = read_list(t, ']', i=i, module=module)
         obj_arr = obj_arr or isinstance(q, (np.ndarray, str))
         arr.append(q)
+        i = skip(t,i,ignore_newline=True)
     if cmatch(t,i,delim):
         i += 1
     try:
@@ -449,17 +500,18 @@ def read_cond(klong, t, i=0):
 
     """
     r = []
-    i,n = klong._expr(t, i)
+    i,n = klong._expr(t, i, ignore_newline=True)
     r.append(n)
     i = cexpect(t, i, ';')
-    i,n = klong._expr(t, i)
+    i,n = klong._expr(t, i, ignore_newline=True)
     r.append(n)
+    i = skip(t,i,ignore_newline=True)
     if cmatch2(t,i,':','|'):
         i,n = read_cond(klong,t,i+2)
         r.append(n)
     else:
         i = cexpect(t, i, ';')
-        i,n = klong._expr(t, i)
+        i,n = klong._expr(t, i, ignore_newline=True)
         r.append(n)
         i = cexpect(t, i, ']')
     return i, KGCond(r)
@@ -469,39 +521,167 @@ def list_to_dict(a):
     return {x[0]:x[1] for x in a}
 
 
-def kg_token(t, i=0, read_neg=False):
+def kg_read(t, i=0, read_neg=False, ignore_newline=False, module=None):
     """
 
-    # Lexeme classes are the sets of the lexemes specified in the
-    # previous section, except for operators.
+        # Lexeme classes are the sets of the lexemes specified in the
+        # previous section, except for operators.
 
-    C := I   # integer
-       | H   # character
-       | R   # real number
-       | S   # string
-       | V   # variable (symbol)
-       | Y   # (quoted) symbol
+        C := I   # integer
+        | H   # character
+        | R   # real number
+        | S   # string
+        | V   # variable (symbol)
+        | Y   # (quoted) symbol
+
+        NOTE: this function mirrors the klong implementation so that sys_read/write
+        match klong's as well.  The grammar read here is a superset of C.
+
+        NOTE: a newline character translates to a semicolon in Klong,
+        except in functions, dictionaries, conditional expressions,
+        and lists. So
+
+        a()
+        b()
+
+        is equal to a();b(), but
+
+        [1
+         2
+         3]
+
+        is equal to [1 2 3] and
+
+        :[x;
+          y;
+          z]
+
+        is equal to :[x;y;z] and
+
+        f::{.d("hello ");
+            .p("world!");
+            []}
+
+        is a valid function definition.
 
     """
-    i = skip(t, i)
+    i = skip(t, i, ignore_newline=ignore_newline)
     if i >= len(t):
         return i, None
     a = t[i]
-    if cmatch2(t, i, '0', 'c'):
+    if a in ['\n', ';']:
+        return i+1,';'
+    elif a in ['(',')','{','}',']']:
+        return i+1,a
+    elif cmatch2(t, i, '0', 'c'):
         return read_char(t, i)
     elif a.isnumeric() or (read_neg and (a == '-' and (i+1) < len(t) and t[i+1].isnumeric())):
         return read_num(t, i)
     elif a == '"':
         return read_string(t, i+1)
     elif a == ':' and (i+1 < len(t)):
-        if t[i+1].isalpha() or t[i+1] == '.':
-            return read_sym(t,i+1)
-        elif t[i+1].isnumeric() or t[i+1] == '"':
-            return kg_token(t, i+1)
-        return i+2,f":{t[i+1]}"
+        aa = t[i+1]
+        if aa.isalpha() or aa == '.':
+            return read_sym(t, i=i+1, module=module)
+        elif aa.isnumeric() or aa == '"':
+            return kg_read(t, i+1, ignore_newline=ignore_newline, module=module)
+        elif aa == '{':
+            i, d = read_list(t, '}', i=i+2, module=module)
+            d = list_to_dict(d)
+            return i, KGCall(KGLambda(lambda x: copy.deepcopy(x)),args=d,arity=0)
+        elif aa == '[':
+            return i+2,':['
+        elif aa == '|':
+            return i+2,':|'
+        return i+2,KGOp(f":{aa}",arity=0)
+    elif safe_eq(a, '['):
+        return read_list(t, ']', i=i+1, module=module)
     elif is_symbolic(a):
-        return read_sym(t, i)
-    return i+1, a
+        return read_sym(t, i, module=module)
+    return read_op(t,i)
+
+
+def kg_write_symbol(x, display=False):
+    return str(x) if display else f":{x}"
+
+
+def kg_write_integer(x, display=False):
+    return str(x)
+
+
+def kg_write_float(x, display=False):
+    return str(x)
+
+
+def kg_write_char(c, display=False):
+    return c if display else f"0c{c}"
+
+
+def kg_write_string(s, display=False):
+    if display:
+        return s
+    arr = ['"']
+    for c in s:
+        if c == '"':
+            arr.append('"')
+        arr.append(c)
+    arr.append('"')
+    return ''.join(arr)
+
+
+def kg_write_dict(d, display=False):
+    arr = [':{']
+    first = True
+    for e in d.items():
+        if not first:
+            arr.append(' ')
+            first = False
+        arr.append(kg_write(e))
+    arr.append('}')
+    return ''.join(arr)
+
+
+def kg_write_list(x, display=False):
+    arr = ['[']
+    for i in range(len(x)):
+        arr.append(kg_write(x[i], display=display))
+        if i < len(x):
+            arr.append(' ')
+    arr.append(']')
+    return ''.join(arr)
+
+
+def kg_write_fn(x, display=False):
+    return str(x)
+
+
+def kg_write_channel(x, display=False):
+    if x.channel_dir == KGChannelDir.INPUT:
+        return ":inchan.0"
+    return f":outchan.{2 if x.raw == sys.stderr else 1}"
+
+
+def kg_write(a, display=False):
+    if isinstance(a,KGSym):
+        return kg_write_symbol(a, display=display)
+    elif is_integer(a):
+        return kg_write_integer(a,display=display)
+    elif is_float(a):
+        return kg_write_float(a,display=display)
+    elif isinstance(a,KGChar):
+        return kg_write_char(a,display=display)
+    elif isinstance(a, str):
+        return kg_write_string(a,display=display)
+    elif isinstance(a,dict):
+        return kg_write_dict(a,display=display)
+    elif is_list(a):
+        return kg_write_list(a,display=display)
+    elif isinstance(a,KGFn):
+        return kg_write_fn(a,display=display)
+    elif isinstance(a,KGChannel):
+        return kg_write_channel(a,display=display)
+    elif safe_eq(a, np.inf):
+        return ":undefined"
 
 
 def kg_argsort(a, descending=False):
@@ -573,6 +753,51 @@ def get_adverb_arity(s, ctx):
 
 
 def merge_projections(arr):
+    """
+
+        A projection is a new function that is created by projecting an
+        existing function onto at least one of its arguments, resulting
+        in the partial application of the original function.
+
+        The notation of projection is that of function application where
+        the arguments onto which the function is being projected are
+        omitted. For instance,
+
+        Projection   Equivalent function
+        {x-y}(5;)    {5-x}
+        {x-y}(;5)    {x-5}
+
+        and, given a ternary function f3:
+
+        Projection   Equivalent function
+        f3(1;2;)     {f3(1;2;x)}
+        f3(1;;3)     {f3(1;x;3)}
+        f3(;2;3)     {f3(x;2;3)}
+        f3(;;3)      {f3(x;y;3)}
+        f3(;2;)      {f3(x;2;y)}
+        f3(1;;)      {f3(1;x;y)}
+
+        The projection of a triad is a dyad or a monad, depending on the
+        number of arguments onto which the triad is being projected. The
+        projection of a dyad is always a monad. There is no projection
+        of a monad or nilad.
+
+        Alternatively, monads and nilads can be considered to be their
+        own projections (onto zero arguments), but there is no special
+        syntax for this case. Any function that is being projected onto
+        all of its arguments is simply the function itself.
+
+        Projections are ordinary functions and can be used in all places
+        where a verb is expected. For instance:
+
+        f::{x,y}
+        f(;0)'[1 2 3]  -->  [[1 0] [2 0] [3 0]]
+        f(0;)'[1 2 3]  -->  [[0 1] [0 2] [0 3]]
+
+        g::{x,y,z}
+        1g(;2;)3       -->  [1 2 3]
+
+    """
     if len(arr) == 0:
         return arr
     if len(arr) == 1 or not has_none(arr[0]):
@@ -595,6 +820,12 @@ def merge_projections(arr):
 
 
 def get_fn_arity(f, level=0):
+    """
+    Examine a function AST and infer arity by looking for x,y and z.
+    This arity is needed to populate the KGFn.
+
+    NOTE: TODO: it maybe easier / better to do this at parse time vs late.
+    """
     if isinstance(f,KGFn):
         x = get_fn_arity(f.a, level=1)
         if is_list(f.args):
