@@ -67,7 +67,7 @@ def eval_dyad_amend(a, b):
                     r[i] = b[0]
         return "".join(["".join(x) for x in r])
     r = np.array(a)
-    if isinstance(b[0],np.ndarray):
+    if is_list(b[0]):
         if len(b[0]) > 1:
             r = r.tolist()
             for i in b[1:]:
@@ -176,11 +176,13 @@ def eval_dyad_at_index(klong, a, b):
     a = str_to_chr_arr(a) if j else a
     if is_list(b):
         if is_empty(b):
-            return a
-        # TODO: return None for missing keys?
-        r = np.asarray([a[x] for x in b])
+            r = np.asarray([])
+        else:
+            # TODO: return None for missing keys? or raise?
+            r = np.asarray([a[x] for x in b])
     elif is_integer(b):
         r = a[b]
+        j = False
     else:
         r = a
     return "".join(r) if j else r
@@ -498,8 +500,7 @@ def eval_dyad_join(a, b):
     aa = a if isinstance(a, list) else ([a[0]] if len(a.shape) > 1 and a.shape[0] == 1 else a.tolist() if len(a.shape) == 1 else [a]) if isinstance(a, np.ndarray) else [a]
     bb = b if isinstance(b, list) else ([b[0]] if len(b.shape) > 1 and b.shape[0] == 1 else b.tolist() if len(b.shape) == 1 else [b]) if isinstance(b, np.ndarray) else [b]
 
-    r = np.asarray([*aa,*bb])
-    return r
+    return np.asarray([*aa,*bb])
 
 
 def eval_dyad_less(a, b):
@@ -560,9 +561,14 @@ def eval_dyad_match(a,b):
                    [1 [2] 3]~[1 [4] 3]  -->  0
 
     """
-    r = vec_fn2(a, b, lambda x,y: (np.isclose(x,y) if (is_number(x) and is_number(y)) else np.all(np.asarray(x,dtype=object) == np.asarray(y,dtype=object))))
-    return kg_truth(rec_flatten(r).all())
-
+    def _e(x,y):
+        return np.isclose(x,y) if (is_number(x) and is_number(y)) else np.all(np.asarray(x,dtype=object) == np.asarray(y,dtype=object))
+    if is_list(a):
+        if is_list(b):
+            return kg_truth(rec_flatten(vec_fn2(a, b, _e)).all() if len(a) == len(b) else 0)
+    elif not is_list(b):
+        return kg_truth(_e(a,b))
+    return False
 
 def eval_dyad_maximum(a, b):
     """
@@ -923,8 +929,8 @@ def create_dyad_functions(klong):
     registry = {}
 
     m = sys.modules[__name__]
-    for x in filter(lambda n: n.startswith("eval_dyad_"), dir(m)):
-        fn = getattr(m,x)
+    for name in filter(lambda n: n.startswith("eval_dyad_"), dir(m)):
+        fn = getattr(m,name)
         name = _get_name(fn.__doc__)
         if fn.__code__.co_argcount == 3:
             fn = lambda x,y,f=fn,klong=klong: f(klong, x, y)
