@@ -1,9 +1,12 @@
 
+import errno
 import os
 import random
 import subprocess
 import sys
-from .core import safe_eq, is_empty
+import time
+
+from .core import KGChannel, KGChannelDir, is_empty, kg_read, kg_write, safe_eq
 
 
 def eval_sys_append_channel(x):
@@ -14,7 +17,7 @@ def eval_sys_append_channel(x):
         See [Output-Channel].
 
     """
-    return open(x, "x")
+    return KGChannel(open(x, "a"), KGChannelDir.OUTPUT)
 
 
 def eval_sys_close_channel(x):
@@ -28,8 +31,8 @@ def eval_sys_close_channel(x):
         current From or To Channel.
 
     """
-    if not x.closed:
-        x.close()
+    if not x.raw.closed:
+        x.raw.close()
 
 
 def eval_sys_display(klong, x):
@@ -40,8 +43,7 @@ def eval_sys_display(klong, x):
         See [Write].
 
     """
-    # TODO: get current channel from klong
-    print(x, end='')
+    klong['.sys.cout'].raw.write(kg_write(x, display=True))
 
 
 def eval_sys_delete_file(x):
@@ -57,8 +59,10 @@ def eval_sys_delete_file(x):
     if os.path.exists(x):
         try:
             os.remove(x)
-        except IOError:
-            raise RuntimeError("file could not be deleted: {x}")
+        except OSError as e:
+            if e.errno != errno.ENOENT:
+                raise RuntimeError("file could not be deleted: {x}")
+            raise e
     else:
         raise RuntimeError("file does not exist: {x}")
 
@@ -73,7 +77,7 @@ def eval_sys_evaluate(klong, x):
         system, e.g. .E("a::123");a will yield 123.
 
     """
-    return klong.exec(x)
+    return klong.exec(x)[-1]
 
 
 def eval_sys_from_channel(klong, x):
@@ -91,7 +95,13 @@ def eval_sys_from_channel(klong, x):
         restore the default From or To Channel (.cin or .cout).
 
     """
-    pass
+    if safe_eq(x,0) or is_empty(x):
+        x = klong['.cin']
+    if x.channel_dir != KGChannelDir.INPUT:
+        raise RuntimeError("output channel cannot be used input")
+    o = klong['.sys.cin']
+    klong['.sys.cin'] = x
+    return o
 
 
 def eval_sys_flush(x):
@@ -103,7 +113,9 @@ def eval_sys_flush(x):
         written to the associated file or device ("flush" the channel).
 
     """
-    x.flush()
+    if x.channel_dir != KGChannelDir.OUTPUT:
+        raise RuntimeError("input channel cannot be flushed")
+    x.raw.flush()
 
 
 def eval_sys_input_channel(x):
@@ -118,14 +130,14 @@ def eval_sys_input_channel(x):
     """
     if os.path.exists(x):
         try:
-            return open(x, 'r')
+            return KGChannel(open(x, 'r'), KGChannelDir.INPUT)
         except IOError:
             raise RuntimeError("file could not be opened: {x}")
     else:
         raise RuntimeError("file does not exist: {x}")
 
 
-def eval_sys_load(x):
+def eval_sys_load(klong, x):
     """
 
         .l(x)                                                     [Load]
@@ -147,10 +159,18 @@ def eval_sys_load(x):
         used to load the value of a single expression from a file.
 
     """
-    pass
+    if os.path.exists(x):
+        try:
+            # TODO: support path defaults as mentioned above
+            with open(x, "r") as f:
+                return klong.exec(f.read())[-1]
+        except IOError:
+            raise RuntimeError("file could not be opened: {x}")
+    else:
+        raise RuntimeError("file does not exist: {x}")
 
 
-def eval_sys_more_input(x):
+def eval_sys_more_input(klong):
     """
 
         .mi(x)                                              [More-Input]
@@ -162,9 +182,10 @@ def eval_sys_more_input(x):
         This is a "negative EOF" function.
 
     """
-    # https://stackoverflow.com/questions/10140281/how-to-find-out-whether-a-file-is-at-its-eof
-    # fh.seek(0, 2); file_size = fh.tell(); fh.seek(0) beforehand and then fh.tell() == file_size
-    pass
+    x = klong['.sys.cin']
+    if x.channel_dir != KGChannelDir.INPUT:
+        raise RuntimeError("output channel cannot be used for input")
+    return not x.at_eof
 
 
 def eval_sys_module(klong, x):
@@ -175,10 +196,13 @@ def eval_sys_module(klong, x):
         Delimit a module. See MODULES documentation.
 
     """
-    pass
+    if safe_eq(x, 0) or is_empty(x):
+        klong.stop_module()
+    else:
+        klong.start_module(x)
 
 
-def eval_sys_output_channel(klong, x):
+def eval_sys_output_channel(x):
     """
 
         .oc(x)                                          [Output-Channel]
@@ -189,7 +213,7 @@ def eval_sys_output_channel(klong, x):
         file and .ac appends to it.
 
     """
-    pass
+    return KGChannel(open(x,"w"), KGChannelDir.OUTPUT)
 
 
 def eval_sys_process_clock(klong):
@@ -208,8 +232,7 @@ def eval_sys_process_clock(klong):
         This function is not avaliable in the Plan 9 port of Klong.
 
     """
-    # time.time - klong.process_start_time()
-    pass
+    return time.time() - klong.process_start_time()
 
 
 def eval_sys_print(klong, x):
@@ -221,8 +244,7 @@ def eval_sys_print(klong, x):
         newline sequence. .p("") will just print a newline.
 
     """
-    # TODO: get current channel
-    print(str(x))
+    klong['.sys.cout'].raw.write(str(x)+"\n")
 
 
 def eval_sys_random_number():
@@ -247,7 +269,16 @@ def eval_sys_read(klong):
         lines.
 
     """
-    pass
+    f = klong['.sys.cin']
+    k = f.raw.tell()
+    r = f.raw.read()
+    if r == '':
+        f.at_eof = True
+        return None
+    else:
+        i,a = kg_read(r,i=0,module=klong.current_module())
+        f.raw.seek(k+i,0)
+        return a
 
 
 def eval_sys_read_line(klong):
@@ -260,7 +291,11 @@ def eval_sys_read_line(klong):
         be stripped from the string.
 
     """
-    pass
+    f = klong['.sys.cin']
+    r = f.raw.readline()
+    if r == '':
+        f.at_eof = True
+    return r.rstrip()
 
 
 def eval_sys_read_string(klong, x):
@@ -274,7 +309,7 @@ def eval_sys_read_string(klong, x):
         forms.
 
     """
-    pass
+    return kg_read(x, i=0, module=klong.current_module())[1]
 
 
 def eval_sys_system(x):
@@ -301,7 +336,13 @@ def eval_sys_to_channel(klong, x):
 
         See [From-Channel]
     """
-    pass
+    if safe_eq(x,0) or is_empty(x):
+        x = klong['.cout']
+    if x.channel_dir != KGChannelDir.OUTPUT:
+        raise RuntimeError("input channel cannot be a to channel")
+    c = klong['.sys.cout']
+    klong['.sys.cout'] = x
+    return c
 
 
 def eval_sys_write(klong, x):
@@ -321,7 +362,7 @@ def eval_sys_write(klong, x):
         ----------------------------------------
         0cx          |  x         | 0cx
         "test"       |  test      | "test"
-        "say ""hi""\"|  say "hi" | "say ""hi""\"
+        "say ""hi""\"|  say "hi"  | "say ""hi""\"
 
         NOTE: the \" was added above to work in docstring
 
@@ -334,8 +375,7 @@ def eval_sys_write(klong, x):
         sequence. Use .p (Print) to do so.
 
     """
-    # TODO: get current channel
-    print(x, end='')
+    klong['.sys.cout'].raw.write(kg_write(x))
 
 
 def eval_sys_exit(x):
